# Configuration
Shotover proxy accepts a two seperate YAML based configuration files. A configuration file specified by `--config-file` 
and a topology file specified by `--topology-file`

## configuration.yaml

The configuration file is used to change general behavior 
of shotover. Currently it supports two values:
* main_log_level
* observability_interface

### main_log_level
This is a single string that you can use to configure logging with shotover. It supports [env_filter](https://docs.rs/env_logger/0.7.1/env_logger/)
style configuration and filtering syntax. Log levels and filters can be dynamically changed while shotover-proxy is still
running.

This can be done by a POST HTTP request to the `/filter` endpoint (configured located at the observability interface) with the 
env_filter string set as the POST data. For example:
```shell script
curl -X PUT -d 'info,shotover_proxy=info' http://127.0.0.1:9001/filter
```

### observability_interface
Defines the interface and port for shotovers observability interface. This interface will serve Prometheus metrics from 
`/metrics` and allow you to configure log levels and filters at `/filter`.
Configured as a string following the Rust SocketAddr parsing rules in the format of `127.0.0.1:8080` for IPV4 addresses or `[2001:db8::1]:8080` for IPV6 addresses. 

## topology.yaml

The topology file is currently the primary method for defining how shotover behaves. Within the topology file you can configure
sources, transforms and transform chains. 

The below documentation shows you what each section does and runs through an entire example of a shotover configuration file.

### `sources` (Sources)
The sources top level resource is a map of named sources, to their definitions.

The sources section of the configuration file allow you to specify a source or origin for requests. 
You can have multiple sources and even multiple sources of the same type. Each is named to allow you to 
easily reference it.

A source will generally represent a database protocol and will accept connections and queries from
a compatible driver. For example the Redis source will accept connections from any Redis (RESP2)
driver (e.g. [redis-py](https://github.com/andymccurdy/redis-py)).

There is a special source type, called a mpsc_chan source (named after the rust multi-producer, single
consumer channel that backs its implementation). This source will only listener to configured topic name
and the associated topic and will then pass the received messages from the channel on it's mapped
transform chain.

There are many `Transforms` that will push a message to a given topic in different ways, enabling complex
asynchronous topologies to be created. 

```yaml
---
# The source section
sources:
  
  # The configured name of the source
  my_named_redis_source:
    # The source type and any configuration needed for it
    # This will generally include a listen address and port
    Redis:
      listen_addr: "127.0.0.1:6379"
  
  # The configured name of the source
  my_cassandra_prod:

    # The source type and any configuration needed for it
    # This will generally include a listen address and port
    Cassandra:
      listen_addr: "127.0.0.1:9042"
      cassandra_ks:
        system.local:
          - key
        test.simple:
          - pk
        test.clustering:
          - pk
          - clustering
  
  # The spcial mpsc_chan source, it will receive messages from a named topic
  mpsc_chan:
    Mpsc:
      topic_name: testtopic
```

### `chain_config` (Chain Configuration)
The `chain_config` top level resource is a map of named chains, to their definitions.

The chain_config section of the configuration file allows you to name and define a transform chain. A transform chain
is represented as an array of transforms and their respective configuration. The order in which a transform chain, is the order
in which a query will traverse it. So the first transform in the chain, will get the request from source first, and pass it
to the second transform in the chain. 

As each transform chain is synchronous, with each transform being able to call the next transform in its chain, the response from the
upstream database or generated by a transform down the chain will be passed back up the chain, allowing each transform to handle the response.

The last transform in a chain should be a "terminating" transform. That is, one that passes the query on to the upstream database
 (e.g. `CassandraCodecDestination`) or one that returns a Response on its own ( e.g. `EchoDestination`).

For example
```yaml
chain_config:
  example_chain:
    - One
    - Two
    - Three
    - TerminatingTransform
``` 

A query from a client will go:

* `Source` -> `One` -> `Two` -> `Three` -> `TerminatingTransform`

The response (returned to the chain by the `TerminatingTransform`) will follow the reverse path:

* `TerminatingTransform` -> `Three` -> `Two` -> `One` -> `Source`

Under the hood, each transform is able to call it's down-chain transform and wait on its response.
Each Transform has its own set of configuration values, options and behavior. See [Transforms](../transforms/transforms.md) for details.

The following example `chain_config` has three chains:

* `redis_chain` - Consists of a MPSCTee, a transform that will copy the query to the named topic and *also* pass the query
down-chain to a terminating transform `RedisDestination` which sends to the query to a Redis server. Very similar to the `tee`
linux program.
* `main_chain` - Also consists of a MPSCTee that will copy queries to the same topic as the `redis_chain` before sending the
query onto caching layer that will try to resolve the query from a redis cache before ending up finally sending the query to
the destination Cassandra cluster via a `CodecDestination`
* `KafkaDestination` - A single transform chain that will serialise and forward any queries to a given Kafka topic.

```yaml
# Chain configuration section
chain_config:

  # The name of the first chain
  my_redis_chain:

    # The first transform in the chain, this case, its the MPSCTee
    - MPSCTee:
        #Configuration for the transform goes here, in this case its the topic to listen on
        topic_name: testtopic

    # The second and final transform for this chain. It will send requests on to the redis server
    # at the remote address
    - RedisDestination:
        remote_address: "127.0.0.1:6378"

  # The second chain, containing three transforms
  my_main_chain:
    - MPSCTee:
        topic_name: testtopic
    - RedisCache:
        config_values: "redis://127.0.0.1/"
    - CodecDestination:
        remote_address: "127.0.0.1:9043"

  # The third chain
  my_async_chain:
    - KafkaDestination:
        config_values:
          bootstrap.servers: "127.0.0.1:9092"
          message.timeout.ms: "5000"
```

### `named_topics` Named Topics
The `named_topics` top level resource is a list of topics. Each topic is backed by a Rust multi-producer, single consumer
channel. 

The value provided is a usize which defines the size of the channel, default 5.

The below example creates a single topic, called `testtopic`. In the example [above](#chain_config-chain-configuration) for `chain_config`, both
the `my_redis_chain` and the `my_main_chain` transform chains would push queries into this topic.

```yaml
named_topics:
  testtopic: 5
```

### `source_to_chain_mapping` Chain Mapping
The `source_to_chain_mapping` top level resource is a map of source names to chain name. This is the binding that will
link a defined source to chain and allow messages/queries generated by a source to traverse a given chain.

The below snippet would complete our entire example:

```yaml
source_to_chain_mapping:
  cassandra_prod: my_main_chain
  redis_prod: my_redis_chain
  mpsc_chan: my_async_chain
```

This mapping would effectively create a solution that:

* All Redis and Cassandra queries get logged to an external kafka cluster. This could be to meet audit or compliance requirements
or even just as the start of an event bus or CQRS based architecture.
* Cassandra queries will be cached and looked up in Redis, before being persisted to the Cassandra cluster. 

The entire example configuration can be found [here](/cass-redis-kafka/config.yaml).